/*
 *  Базовый скетч для платы WeMoS D1 для взаимодействия с платформой ThingWorx
 *  Рабочий инстанс:  tvsn0428.cloud.thingworx.com 
 *  
 *  Скетч использует EEPROM платы ESP8266 для хранения параметров сети и сервисов начиная с нулевого адреса (Смещение EEPROMOffset)
 *  Первые 4 байта хранят символы TVSN
 *  Следующие 4 байта хранят номер версии начиная с 0001 - имеется ввиду номер версии таблицы размещения информации 
 *  and REDGREEN on D5
 *  and REDBLUE on D4
 *  You need 
 *
 *  Процедуры чтения-записи с консоли
 *    Терминал должен быть настроен на отправку конца строки \n либо для символа конца строки используется символ "#" (шарп, диез)
 *
 *
 */

#include <ESP8266WiFi.h>
#include <EEPROM.h>
#include <WiFiClientSecure.h>
#include <ArduinoJson.h>

// отладочные переменные
#define Debug_ShowEEPROMParameters true  // выводить для просмотра значения параметров, сохраненные / сохраняемые в EEPROM
#define Debug_ShowPrivateParameters true  // выводить скрытую информацию, такую как пароли и ключи
#define Defalut_BeforeScanNetwork true // проводить предварительное сканирование WiFi сетей

// базовые настройки скетча по-умолчанию
#define DefaultWiFiSSID "SmartThingsNet" // <-- Имя WiFi сети по умолчанию         // для настройки телефона
#define DefaultWiFiPswd "12345678" // <-- Ключ WiFi сети по умолчанию

#define DefaultThingWorxThing "PolytechThing0000";                        // Вещь
#define DefaultThingWorxService "RemoteTerminal";                         // Сервис
#define DefaultThingWorxAppKey  "4ab21f71-8e02-47d5-93be-b1a7cf28ae76";   // Ключ

/*
Вещь: ISThing1, Сервис: InputService, Ключ: 873589e0-35f4-4b66-a7da-8a9fd2d49762
Вещь: ISThing2, Сервис: InputService, Ключ: 0de62d28-e61a-46ae-8b69-6edcd41df6c9
Вещь: ISThing3, Сервис: InputService, Ключ: 9d4cba28-3361-48c9-bfbf-a9f595fb6459
Вещь: ISThing4, Сервис: InputService, Ключ: 8ff4eb1c-0c9a-48f5-8031-affb1ef23f86
*/

#define EEPROMSize 512  // общий объем EEPROM
#define EEPROMMaxLineLength 63 // максимальная длина строки при сохранении / чтении в EEPROM - не забываем о '\0' в конце
#define ConsoleBufferLength 64  // количество символов в буфере в ОЗУ при работе с консолью

// Таблица форматирования EEPROM - размещения информационных полей, а также контрольные ключи
// Ключ опознавания формата данных таблицы
#define EEPROMSmartAgroROMLabel "SmartAgroCB001" // контрольная панель - версия 001
#define EEPROMSmartAgroROMLabel_Start 0 // размещаем с самого начала памяти
#define EEPROMSmartAgroROMLabel_Length 15 // длина текста
// Настройки WiFi
#define EEPROMWIFISSID_Start 16  // с 17 байта идёт идентификатор известной сети
#define EEPROMWIFISSID_Length 15 // длина идентификатора известной сети
#define EEPROMWIFIPswd_Start 32 // с 33 байта идет ключ сети
#define EEPROMWIFIPswd_Length 15 // длина ключа сети
// Настройки доступа к облаку
#define EEPROMThing_Start 48
#define EEPROMThing_Length 31
#define EEPROMService_Start 80
#define EEPROMService_Length 31
#define EEPROMAppKey_Start 112
#define EEPROMAppKey_Length 63

// Определения для работы с консолью
#define ConsoleBufferLength 64  // количество символов в буфере
#define showConsoleEcho true    // неоходимость отображения (эха) введенных символов

#define RequestDelay 5000 // Задержка при отправке запросов (5 секунд) ------------------------------

// Переменные для связи
String LSSID = DefaultWiFiSSID;
String LPswd = DefaultWiFiPswd;

String host = "tvsn0428.cloud.thingworx.com";
const char* hostlink = "tvsn0428.cloud.thingworx.com";

String thingworxthingapx = "/Thingworx/Things/";
String thingworxservicesapx = "/Services/";
String thingworxmethodapx = "?method=post&appKey=";
String thingworxcontrolapx = "&x-thingworx-session=true<>";


String thingworxThing = DefaultThingWorxThing;
String thingworxService = DefaultThingWorxService;
String thingworxAppKey = DefaultThingWorxAppKey;

// Максимальное время ожидания ответа от сервера
#define IOT_TIMEOUT1 5000
#define IOT_TIMEOUT2 100

// Таймер ожидания прихода символов с сервера
long timer_iot_timeout = 0;

// Размер приемного буффера
#define BUFF_LENGTH 512

// Приемный буфер
char buff[BUFF_LENGTH] = "";

String text;

void setup() {
    // Start Serial
    Serial.begin(9600);
    delay(10);

    SerialPort_ClearInternalBuffer(); // очистка буфера COM-порта от "мусорных" данных

    EEPROM.begin(512); // <-- Только для WEMOS D1

    // Check EEPROM format  
    String LText = EEPROM_ReadString(EEPROMSmartAgroROMLabel_Start, EEPROMSmartAgroROMLabel_Length);
    Serial.println();
    Serial.println(LText);
    if (LText != EEPROMSmartAgroROMLabel)  // плата не "прописана", выводим запрос на "инициализацию" EEPROM и ждём  
    {
        Serial.println();
        Serial.println("The board is not activated for SmartAgro. Do you whant activate this (y / n)? All EEPROM datas will erase.");
        char lcmd = ' ';
        lcmd = SerialPort_ReadKey(); 
        if ((lcmd != 'y')&&(lcmd != 'Y'))
        {
            // любая другая клавиша - зависаем систему с сообщением о необходимости перезагрузки
            Serial.println("Abort activation.");
            FreezeBoard();        
        } 
        // "Прописываем в плату заголовок и данные по умолчанию
        LText = EEPROMSmartAgroROMLabel;
        EEPROM_ReWriteString(LText, EEPROMSmartAgroROMLabel_Start, EEPROMSmartAgroROMLabel_Length);
        WriteDefaultWiFiDataToEEPROM(); // записать данные по умолчанию
        WriteDefaultThingWorxDataToEEPROM(); // записать данные по умолчанию
    }
    ReadWiFiDataFromEEPROM(); // читаем все данные из долговременной памяти
    ReadThingWorxDataFromEEPROM(); // читаем все данные из долговременной памяти
    // Проверяем считанные данные на допустимость
    bool lctrl = true;
    //
    //
    //
    //
    // Startup timeout
    bool isUserNeedSetup = StartUpAnyKeyWaithing(5); // <-- ожидание 5 секунд на возможность настройки
    if (isUserNeedSetup)
    {
        SetupEEPROMTableDialog(); // настроечный диалог 
    }

 
  // We start by connecting to a WiFi network

  Serial.println();
  Serial.println();
    
  WiFi.mode(WIFI_STA);
  WiFi.disconnect();
  delay(100);
  
  if (Defalut_BeforeScanNetwork)  // предварительное сканирование Wifi-сетей
  {
      // Scan networks
      Serial.println("scan start");
      // WiFi.scanNetworks will return the number of networks found
      int n = WiFi.scanNetworks();
      Serial.println("scan done");
      if (n == 0)
        Serial.println("no networks found");
      else
      {
        Serial.print(n);
        Serial.println(" networks found");
        for (int i = 0; i < n; ++i)
        {
          // Print SSID and RSSI for each network found
          Serial.print(i + 1);
          Serial.print(": ");
          Serial.print(WiFi.SSID(i));
          Serial.print(" (");
          Serial.print(WiFi.RSSI(i));
          Serial.print(")");
          Serial.println((WiFi.encryptionType(i) == ENC_TYPE_NONE)?" ":"*");
          delay(10);
        }
      }
      Serial.println("");
      WiFi.disconnect();
      delay(500);
  }  // if   Defalut_BeforeScanNetwork
  
  //WiFi.begin(ssid, password);
  Serial.println();
  Serial.println("=============");
  Serial.println(LSSID);
  Serial.println(LPswd);
  Serial.println("=============");
  char* cssid;
  if (LSSID.length()!=0)
  {
      cssid = const_cast<char*>(LSSID.c_str());
  }
  //cssid = string2char(LSSID);
  //LSSID.toCharArray(cssid, LSSID.length());
  char* cpassword;
  if (LPswd.length()!=0)
  {
      cpassword = const_cast<char*>(LPswd.c_str());
  }
  //cpassword = string2char(LPswd);
  //LPswd.toCharArray(cpassword, LPswd.length());

  // Connect to known network
  Serial.print("Connecting to ");
  Serial.println(cssid);
  if (Debug_ShowPrivateParameters)  // Если выводить приватную информацию
  {
      Serial.println(cpassword);  // <-- приватная информация
  }
  
  WiFi.begin(cssid, cpassword);

//  wifi_station_clear_cert_key();
  
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }

  Serial.println("");
  Serial.println("WiFi connected");  
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());

  pinMode(D7, INPUT);   // вход 
  pinMode(D6, INPUT);   // вход
  pinMode(D5, INPUT);   // вход
  pinMode(D4, INPUT);   // вход
  pinMode(D3, OUTPUT);  // выход
  pinMode(D2, OUTPUT);  // выход
  pinMode(D1, OUTPUT);  // выход
  pinMode(D0, OUTPUT);  // выход
} // ========================== end of setup() =======================

//
long cycleid = 0; // номер цикла от начала 
int value = 0;
int analogValue = 0;  // аналоговый пин
int digitalValues[8] = {0,0,0,0,0,0,0,0}; // цифровые пины
// параметры "от платы" - a0, d7, d6, d5, d4
// параметры "к плате" - D3, D2, D1, D0
int ledvalue = 0;


void loop() {  // ================ основной цикл ======================

    delay(RequestDelay);

    cycleid++; // for debugging

    // Читаем данные из пинов
    analogValue = analogRead(A0);
    digitalValues[7] = digitalRead(D7);
    digitalValues[6] = digitalRead(D6);
    digitalValues[5] = digitalRead(D5);
    digitalValues[4] = digitalRead(D4);
    Serial.println();
    Serial.println();
    Serial.print("connecting to ");
    Serial.println(host);
  
    // Use WiFiClient class to create TCP connections
    //WiFiClient client;
    // Use WiFiClientSecure class to create TLS connection
    WiFiClientSecure client;
    const int httpPort = 443; //80;
    if (!client.connect(hostlink, httpPort)) {
        Serial.println("connection failed");
        return;
    }
  
    String htmlbody = "{\"a0\":\"" + String(analogValue) + "\", ";
    htmlbody += "\"d7\":\"" + String(digitalValues[7]) + "\", ";
    htmlbody += "\"d6\":\"" + String(digitalValues[6]) + "\", ";
    htmlbody += "\"d5\":\"" + String(digitalValues[5]) + "\", ";
    htmlbody += "\"d4\":\"" + String(digitalValues[4]) + "\"}";   
    
    int htmlbodylength = htmlbody.length();

    // We now create a URI for the request
    String reqline = "POST " + thingworxthingapx + thingworxThing + thingworxservicesapx + thingworxService;
    reqline += thingworxmethodapx + thingworxAppKey + thingworxcontrolapx + " HTTP/1.1\n";
    reqline += "Host: " + host + "\n";
    reqline += "Accept: application/json\n";
    reqline += "Content-Type: application/json\n";
    reqline += "Content-Length: " + String(htmlbodylength) + "\n\n";
    reqline += htmlbody + "\n";

    // пытаемся соединиться и отправить данные

    Serial.println("Connecting to IoT server...");
    Serial.println("Request:");
    Serial.println(reqline);
  
    // This will send the request to the server
    client.print(reqline);
  
    // Ждем ответа от сервера
    timer_iot_timeout = millis();
    while ((client.available() == 0) && (millis() < timer_iot_timeout + IOT_TIMEOUT1));

    // Выводим ответ о сервера, и, если медленное соединение, ждем выход по таймауту
    int iii = 0; // эквивалентно очистке буфера buff[iii]
    bool currentLineIsBlank = true;
    bool flagJSON = false;
    text = "";
    timer_iot_timeout = millis();
    while ((millis() < timer_iot_timeout + IOT_TIMEOUT2) && (client.connected()))
    {
        while (client.available() > 0)
        {
            char symb = client.read();
            Serial.print(symb);
            if (symb == '{')
            {
                flagJSON = true;
                text = "";
            }
            else if (symb == '}')
            {            
                flagJSON = false;
            }
            if (flagJSON == true)
            {
                if (iii < BUFF_LENGTH - 2)  // плюс фигурная скобка и символ конца строки
                {
                    buff[iii] = symb;
                    iii ++;
                    text += symb;
                }
            }
            timer_iot_timeout = millis();
        }
    }
    buff[iii] = '}';
    buff[iii + 1] = '\0';
    text+='}';
    Serial.println(buff);
    Serial.println(text);
    // Закрываем соединение
    client.stop();
  
    Serial.println();
    Serial.println("closing connection");

    // --- Расшифровываем параметры ----

    StaticJsonBuffer<BUFF_LENGTH> jsonBuffer;
    JsonObject& json_array = jsonBuffer.parseObject(buff);
    digitalValues[0] = json_array["D0"];            
    digitalValues[1] = json_array["D1"];            
    digitalValues[2] = json_array["D2"];                  
    digitalValues[3] = json_array["D3"];      
    Serial.println();
    Serial.println("Anwser parameters:");
    Serial.print("D0: " + String(digitalValues[0]) + "   ");
    Serial.print("D1: " + String(digitalValues[1]) + "   ");
    Serial.print("D2: " + String(digitalValues[2]) + "   ");
    Serial.print("D3: " + String(digitalValues[3]));      

    // --- Обрабатываем пришедшие параметры -----
    
    if (digitalValues[0] == 1)  {  digitalWrite(D0, HIGH); } else  {  digitalWrite(D0, LOW); }
    if (digitalValues[1] == 1)  {  digitalWrite(D1, HIGH); } else  {  digitalWrite(D1, LOW); }
    if (digitalValues[2] == 1)  {  digitalWrite(D2, HIGH); } else  {  digitalWrite(D2, LOW); }
    if (digitalValues[3] == 1)  {  digitalWrite(D3, HIGH); } else  {  digitalWrite(D3, LOW); }
        
}

// ===================== Функциональность стартового меню настроек =======

boolean StartUpAnyKeyWaithing(int TimeOutLength) // ожидание нажатия клавиши
{
    boolean ctrl = false;
    char inputChar;
    if (TimeOutLength > 0)
    {
        Serial.println();
        Serial.print("Press any key for setup or wait ");
        Serial.print(TimeOutLength);
        Serial.println(" sec for start");
        for (int i = 0; i < TimeOutLength + 1; i++)
        {
            Serial.print(".");
            if (Serial.available())
            {
                inputChar = Serial.read();
                Serial.println(inputChar);
                Serial.println("Enter setup");
                ctrl = true;
                break;
            }
            delay(1000);
        } // for        
    } // if TimeOutLength
    SerialPort_ClearInternalBuffer(); // очистка буфера COM-порта от пришедших данных
    return ctrl;   
}


void ReadWiFiDataFromEEPROM()  // считать все данные из EEPROM
{  
    LSSID = EEPROM_ReadString(EEPROMWIFISSID_Start, EEPROMWIFISSID_Length);
    LPswd = EEPROM_ReadString(EEPROMWIFIPswd_Start, EEPROMWIFIPswd_Length);
} // --- end of ReadWiFiDataFromEEPROM() ----

void ReadThingWorxDataFromEEPROM()  // считать все данные из EEPROM
{  
    thingworxThing = EEPROM_ReadString(EEPROMThing_Start, EEPROMThing_Length);
    thingworxService = EEPROM_ReadString(EEPROMService_Start, EEPROMService_Length);
    thingworxAppKey = EEPROM_ReadString(EEPROMAppKey_Start, EEPROMAppKey_Length);
} // --- end of ReadThingWorxDataFromEEPROM() ----

void WriteDefaultWiFiDataToEEPROM() // записать данные по умолчанию 
{
    String sdata = DefaultWiFiSSID;
    EEPROM_ReWriteString(sdata, EEPROMWIFISSID_Start, EEPROMWIFISSID_Length);
    sdata = DefaultWiFiPswd; 
    EEPROM_ReWriteString(sdata, EEPROMWIFIPswd_Start, EEPROMWIFIPswd_Length);
} // --- end of WriteDefaultWiFiDataToEEPROM() -----

void WriteDefaultThingWorxDataToEEPROM() // записать данные по умолчанию 
{
    String sdata = DefaultThingWorxThing;
    EEPROM_ReWriteString(sdata, EEPROMThing_Start, EEPROMThing_Length);
    sdata = DefaultThingWorxService;
    EEPROM_ReWriteString(sdata, EEPROMService_Start, EEPROMService_Length);
    sdata = DefaultThingWorxAppKey;
    EEPROM_ReWriteString(sdata, EEPROMAppKey_Start, EEPROMAppKey_Length);
} // --- end of WriteDefaultThinWorxDataToEEPROM() -----

bool CheckValidParameterLine(String AString) // проверка допустимости строки (цифры, лат.символы и дефис)
{
    bool lctrl = true;
    int n = AString.length();
    char lchar;
    for (int i = 0; i < n; i++)
    {
        lchar = AString[i];
        //if (!(
    } // for

    return lctrl;
}

void DrawSerialMenu()
{
    Serial.println();
    Serial.println("*******************************");
    Serial.println("*********  SETUP  MENU  *******");
    Serial.println("*******************************");
    if (Debug_ShowEEPROMParameters) 
    {
        Serial.println("? - Show current parameters");
    }
    Serial.println("1 - Set Wifi network SSID");
    Serial.println("2 - Set Wifi network key");
    Serial.println("3 - Set ThingWorx Thing name");
    Serial.println("4 - Set ThingWorx Service name");
    Serial.println("5 - Set ThingWorx AppKey name");
    Serial.println("6 - Reload default WiFi settings");    
    Serial.println("7 - Reload default ThingWorx settings");
    Serial.println("0 - Exit the menu");
    Serial.println("*******************************");
}

void SetupEEPROMTableDialog() // настроечный диалог, с отображением меню и т.п. 
{
    char ch;
    char lcmd;
    bool ctrl = true;
    String s;
    SerialPort_ClearInternalBuffer(); // очистка буфера COM-порта от "мусорных" данных
    do
    {
        DrawSerialMenu();
        ch = SerialPort_ReadKey();
        switch (ch)
        {
            case '?':
                if (Debug_ShowEEPROMParameters) 
                {
                    Serial.print("SSID: ");
                    Serial.println(LSSID);
                    Serial.print("SKey: ");
                    Serial.println(LPswd);
                    Serial.print("Thing name: ");
                    Serial.println(thingworxThing);
                    Serial.print("Service name: ");
                    Serial.println(thingworxService);
                    Serial.print("Appkey: ");
                    Serial.println(thingworxAppKey);
                }
                break; 
            case '1':
                Serial.println("Please, Enter Wifi network SSID (# - end of line)");
                s = SerialPort_ReadLine();
                // < -- здесь хорошо бы проверить строку на допустимость
                EEPROM_ReWriteString(s, EEPROMWIFISSID_Start, EEPROMWIFISSID_Length);  
                LSSID = EEPROM_ReadString(EEPROMWIFISSID_Start, EEPROMWIFISSID_Length);
                // отладочный режим
                if (Debug_ShowEEPROMParameters) 
                {
                    Serial.print("SSID: ");
                    Serial.println(LSSID);
                }
                break;
            case '2':  
                Serial.println("Please, Enter Wifi network key (# - end of line)");
                s = SerialPort_ReadLine();
                // < -- здесь хорошо бы проверить строку на допустимость
                EEPROM_ReWriteString(s, EEPROMWIFIPswd_Start, EEPROMWIFIPswd_Length);  
                LPswd = EEPROM_ReadString(EEPROMWIFIPswd_Start, EEPROMWIFIPswd_Length);
                // отладочный режим
                if (Debug_ShowEEPROMParameters) 
                {
                    Serial.print("SKey: ");
                    Serial.println(LPswd);
                }
                break;
            case '3':  
                Serial.println("Please, Enter Thing name (# - end of line)");
                s = SerialPort_ReadLine();
                // < -- здесь хорошо бы проверить строку на допустимость
                EEPROM_ReWriteString(s, EEPROMThing_Start, EEPROMThing_Length);  
                thingworxThing = EEPROM_ReadString(EEPROMThing_Start, EEPROMThing_Length);
                // отладочный режим
                if (Debug_ShowEEPROMParameters) 
                {
                    Serial.print("Thing name: ");
                    Serial.println(thingworxThing);
                }
                break;
            case '4':  
                Serial.println("Please, Enter Service name (# - end of line)");
                s = SerialPort_ReadLine();
                // < -- здесь хорошо бы проверить строку на допустимость
                EEPROM_ReWriteString(s, EEPROMService_Start, EEPROMService_Length);  
                thingworxService = EEPROM_ReadString(EEPROMService_Start, EEPROMService_Length);
                // отладочный режим
                if (Debug_ShowEEPROMParameters) 
                {
                    Serial.print("Service name: ");
                    Serial.println(thingworxService);
                }
                break;
            case '5':  
                Serial.println("Please, Enter Appkey (# - end of line)");
                s = SerialPort_ReadLine();
                // < -- здесь хорошо бы проверить строку на допустимость
                EEPROM_ReWriteString(s, EEPROMAppKey_Start, EEPROMAppKey_Length);  
                thingworxAppKey = EEPROM_ReadString(EEPROMAppKey_Start, EEPROMAppKey_Length);
                // отладочный режим
                if (Debug_ShowEEPROMParameters) 
                {
                    Serial.print("Appkey: ");
                    Serial.println(thingworxAppKey);
                }
                break;
            case '6': 
                WriteDefaultWiFiDataToEEPROM(); // записать данные по умолчанию
                ReadWiFiDataFromEEPROM(); // читаем все данные из долговременной памяти
                break;
            case '7':  
                WriteDefaultThingWorxDataToEEPROM(); // записать данные по умолчанию
                ReadThingWorxDataFromEEPROM(); // читаем все данные из долговременной памяти             
                break;
            case '0':  
                Serial.println("Exit from menu (y / n)?");
                lcmd = SerialPort_ReadKey(); 
                if (!((lcmd == 'y')||(lcmd == 'Y')))
                {
                    ch = '-';
                }
                break;
        } // switch
    } while (ch != '0'); 
}

// ===================== Общеслужебные функции ===========================

void FreezeBoard()  // Зависаем плату с сообщением о необходимости перезагрузки
{
    Serial.println("The board stopped. Please reset the board.");
    while (true)
    {
        // Do nothing
    } // while
}

// ===================== Функции для работы со строками =================

char* string2char(String command){
    if(command.length()!=0){
        char *p = const_cast<char*>(command.c_str());
        return p;
    }
}

char* ExString2Char(String cmd)
{
    String stringVar=cmd;
    char charVar[sizeof(stringVar)];
    stringVar.toCharArray(charVar, sizeof(charVar));
    
}

// ===================== Функции для работы с Serial-консолью ============

void SerialPort_ClearInternalBuffer() // очистка буфера COM-порта от пришедших данных
{
    char ch;
    while (Serial.available())
    {
        ch = Serial.read();
    }
}

bool SerialPort_KeyPressed()  // "проверяем нажатие" - фактически, наличие символа в буфере
{
    bool ctrl = false;
    if (Serial.available())
    {
       ctrl = true;
    }     
    return ctrl;
}

char SerialPort_KeyPressedEx()  // "проверяем нажатие" - фактически, наличие символа в буфере и возращаем его или \0
{
    char ch = '\0';
    bool ctrl = false;
    if (Serial.available())
    {
        ch = Serial.read();
        ctrl = true;
    }     
    return ch;
}

char SerialPort_ReadKey()  // дожидаемся ввода символа и возвращаем его
{
    char ch = '\0';
    bool ctrl = true;     
    while (ctrl)
    {
        if(Serial.available())
        {
            ch = Serial.read();
            if (showConsoleEcho)
            {
                Serial.print(ch);
            }
            ctrl = false;
        }
    }     
    return ch;
}

String SerialPort_ReadLine()
{
    String inputString;
    bool ctrl = true;
    char inputChar = '\0';
    while (ctrl)
    {
        if(Serial.available())
        {
            inputChar = Serial.read();
            if (showConsoleEcho)
            {
                Serial.print(inputChar);
            }
            // Обработка символов
            if ((inputChar == '\n')||(inputChar == '#')) // <-- команда закрывается концом строки или символом шарпа ---
            {
                ctrl = false;
            }
            else if (inputString.length() < ConsoleBufferLength - 1)  // если длина буфера не превышена
            {
                inputString += inputChar;
            }
        }
    } 
    return inputString;
}

String SerialPort_ReadLineEx()  // функция с возможностью отмены ввода. При отмене возвращает \0
{
    String inputString;
    bool ctrl = true;
    char inputChar = '\0';
    while (ctrl)
    {
        if(Serial.available())
        {
            inputChar = Serial.read();
            if (showConsoleEcho)
            {
                Serial.print(inputChar);
            }
            // Обработка символов
            if ((inputChar == '\n')||(inputChar == '#')) // <-- команда закрывается концом строки или символом шарпа ---
            {
                ctrl = false;
            }
            else if ((inputChar == 27) || (inputChar == 3)) // код клавиш ESC и сочетания Ctrl + C
            {
                inputString = "\0";
                ctrl = false;
            }
            else if (inputString.length() < ConsoleBufferLength - 1)  // если длина буфера не превышена
            {
                inputString += inputChar;
            }
        }
    } 
    return inputString;
}

String SerialPort_Prompt(String AMessage, String ADefaultValue)
{
    String inputString;
    Serial.println(AMessage);
    inputString = SerialPort_ReadLineEx();
    if (inputString == "\0") 
    {
        inputString = ADefaultValue;
    }
    return inputString;
}

// ===================== Функции для работы с EEPROM =====================

void EEPROM_WriteString(String AString, int AStart)
{
    char LChar;
    byte LValue;
    int n = AString.length();
    if ( n > EEPROMMaxLineLength)
    {
        n = EEPROMMaxLineLength;  // <--- "обрезаем" слишком длинные строки      
    }
    if ((AStart >= 0) && (AStart + n + 1 < EEPROMSize )) // если строка + '\0' лезут в стираемую память
    {
        for (int i = 0; i < n; i++)
        {
            LChar = AString[i];
            LValue = LChar;
            EEPROM.write(AStart + i, LValue);
        }
        // Теперь дописываем NULL в конец строки
        LValue = 0;
        EEPROM.write(AStart + n, LValue);
        EEPROM.commit(); // < --- Только для WeMoS D1 - подтверждаем запись в EEPROM
    } // 
} // --- end of EEPROM_WriteString() ------------

void EEPROM_ReWriteString(String AString, int AStart, int AMaxLength) // Перезапись строки в EEPROM с затиранием всего блока AMaxLength
{
    char LChar;
    byte LByte;
    int LMaxLength = AMaxLength;    
    int n = AString.length();
    if (LMaxLength > EEPROMMaxLineLength)
    {
        LMaxLength = EEPROMMaxLineLength; // <--- "обрезаем" слишком длинный буфер
    }
    if ((AStart >= 0) && (LMaxLength > 0) && (AStart + LMaxLength < EEPROMSize))
    {
        if (n > LMaxLength - 1)
        {
            n = LMaxLength - 1; // длина слишком большая, обрезаем. Плюс оставляем место для NULL
        }
        for (int i = 0; i < LMaxLength; i++) // копируем символя строки буфера, затирая остальные \0
        {
            if (i < n)  // записываем символ из строки
            {
                LChar = AString[i];
                LByte = LChar; 
                EEPROM.write(AStart + i, LByte);                            
            }
            else        // записываем NULL
            {
                EEPROM.write(AStart + i, 0);
            }
        }  // for i
        
        EEPROM.commit(); // < --- Только для WeMoS D1 - подтверждаем запись в EEPROM    
    }
} // --- end of EEPROM_ReWriteString() -------------

String EEPROM_ReadString(int AStart, int AMaxLength)  // чтение строки из EEPROM 
{
    String LString = "";
    char LChar = ' ';
    byte LByte = 0;
    int LMaxLength = AMaxLength;
    if (LMaxLength > EEPROMMaxLineLength)
    {
        LMaxLength = EEPROMMaxLineLength; // <--- "обрезаем" слишком длинный буфер
    }
    if ((AStart >= 0) && (LMaxLength > 0) && (AStart + LMaxLength < EEPROMSize))
    {
        for (int i = 0; i < LMaxLength; i++)
        {
            LByte = EEPROM.read(AStart + i);
            if ((LByte != 0) && (LByte != 255))
            {
                LChar = LByte;
                LString += LChar;
            }
            else
            {
                break;
            }
        }  // for
    } // for    
    return LString;
} // -- end of EEPROM_ReadString() ------------

